__all__ = ['suggest_filename', 'template_to_filepath']

import os
import re

from .constants import CHARACTER_REPLACEMENTS, TEMPLATE_PATTERNS
from .utils import list_to_single_value


def suggest_filename(metadata):
	"""Generate a filename like Google for a song based on metadata.

	Parameters:
		metadata (dict): A metadata dict.

	Returns:
		str: A filename string without an extension.
	"""

	metadata = {k: list_to_single_value(v) for k, v in metadata.items()}

	if 'title' in metadata and 'track_number' in metadata:  # Music Manager.
		suggested_filename = f"{metadata['track_number']:0>2} {metadata['title']}"
	elif 'title' in metadata and 'trackNumber' in metadata:  # Mobile.
		suggested_filename = f"{metadata['trackNumber']:0>2} {metadata['title']}"
	elif 'title' in metadata and 'tracknumber' in metadata:  # Mutagen.
		suggested_filename = f"{metadata['tracknumber']:0>2} {metadata['title']}"
	else:
		suggested_filename = f"00 {metadata.get('title', '')}"

	return suggested_filename


def _split_track_number(field):
	match = re.match(r'(\d+)(?:/\d+)?', field)

	return match.group(1) if match else field


# TODO: Revisit template_to_filepath.
def _replace_template_patterns(template, metadata, template_patterns):
	drive, path = os.path.splitdrive(template)
	parts = []

	while True:
		newpath, tail = os.path.split(path)

		if newpath == path:
			break

		parts.append(tail)
		path = newpath

	parts.reverse()

	for i, part in enumerate(parts):
		for key in template_patterns:
			if key in part and template_patterns[key] in metadata:
				# Force track number to be zero-padded to 2 digits.
				if any(
					template_patterns[key] == tracknumber_field for tracknumber_field in ['tracknumber', 'track_number']
				):
					track_number = list_to_single_value(_split_track_number(str(metadata[template_patterns[key]])))
					metadata[template_patterns[key]] = track_number.zfill(2)

				parts[i] = parts[i].replace(key, metadata[template_patterns[key]])

		for char in CHARACTER_REPLACEMENTS:
			if char in parts[i]:
				parts[i] = parts[i].replace(char, CHARACTER_REPLACEMENTS[char])

	if os.path.isabs(template):
		if drive:
			filepath = os.path.join(drive, os.sep, *parts)
		else:
			filepath = os.path.join(os.sep, *parts)
	else:
		filepath = os.path.join(*parts)

	return filepath


def template_to_filepath(template, metadata, template_patterns=None):
	"""Create directory structure and file name based on metadata template.

	Parameters:
		template (str): A filepath which can include template patterns as defined by :param template_patterns:.

		metadata (dict): A metadata dict.

		template_patterns (dict): A dict of ``pattern: field`` pairs used to replace patterns with metadata field values.
			Default: :const:`~google_music_utils.constants.TEMPLATE_PATTERNS`

	Returns:
		str: A filepath.
	"""

	if template_patterns is None:
		template_patterns = TEMPLATE_PATTERNS

	metadata = {k: list_to_single_value(v) for k, v in metadata.items()}
	assert isinstance(metadata, dict)

	suggested_filename = suggest_filename(metadata)

	if template == os.getcwd() or template == '%suggested%':
		filepath = suggested_filename
	else:
		t = template.replace('%suggested%', suggested_filename)
		filepath = _replace_template_patterns(t, metadata, template_patterns)

	return filepath



from .__about__ import *
from .compare import *
from .constants import *
from .filter import *
from .metadata import *
from .misc import *

__all__ = [
	*__about__.__all__,
	*compare.__all__,
	*constants.__all__,
	*filter.__all__,
	*metadata.__all__,
	*misc.__all__
]



__all__ = ['compare_item_collections']

import audio_metadata
from multidict import MultiDict

from .constants import FIELD_MAP
from .utils import list_to_single_value, normalize_value


def _gather_field_values(item, *, fields=None, field_map=FIELD_MAP, normalize_values=False):
	"""Create a tuple of normalized metadata field values.

	Parameter:
		item (dict): Item dict or filepath.
		fields (list): A sequence of fields used to compare item dicts.
		normalize_values (bool): Normalize metadata values to remove common differences between sources.
			Default: ``False``

	Returns:
		tuple: Values from the given metadata fields.
	"""

	if not isinstance(item, (audio_metadata.Format, dict)):
		item = audio_metadata.load(item).tags  # pragma: no cover

	if fields is None:
		if hasattr(item, 'FIELD_MAP'):
			fields = [item.FIELD_MAP.get(k, k) for k in item]
		else:
			fields = list(item.keys())

	normalize = normalize_value if normalize_values else lambda x: str(x)

	field_values = []

	for field in fields:
		if item.get(field):
			field_values.append(normalize(list_to_single_value(item[field])))
		elif isinstance(field_map, MultiDict):
			for alias in field_map.getall(field, []):
				if item.get(alias):  # pragma: no branch
					field_values.append(normalize(list_to_single_value(item[alias])))
					break
		elif isinstance(field_map, dict):  # pragma: no branch
			alias = field_map.get(field)

			if alias in item:  # pragma: no branch
				field_values.append(normalize(list_to_single_value(item[alias])))

	return tuple(field_values)


def compare_item_collections(src, dst, *, fields=None, field_map=None, normalize_values=False):
	"""Find items from an item collection that are not in another item collection.

	Parameters:
		src (list): A sequence of item dicts or filepaths.
		dst (list): A sequence of item dicts or filepaths.
		fields (list): A sequence of fields used to compare item dicts.
		normalize_values (bool): Normalize metadata values to remove common differences between sources.
			Default: ``False``

	Yields:
		dict: The next item from ``src`` collection not in ``dst`` collection.

	Example:
		>>> from google_music_utils import compare_item_collections
		>>> list(compare_item_collections(song_list_1, song_list_2, fields=['artist', 'album', 'title']))
	"""

	if field_map is None:
		field_map = FIELD_MAP

	dst_keys = {
		_gather_field_values(
			dst_item, fields=fields, field_map=field_map, normalize_values=normalize_values
		) for dst_item in dst
	}

	for src_item in src:
		if _gather_field_values(
			src_item, fields=fields, field_map=field_map, normalize_values=normalize_values
		) not in dst_keys:
			yield src_item



__all__ = [
	'__author__', '__author_email__',
	'__copyright__', '__license__',
	'__summary__', '__title__',
	'__url__', '__version__',
	'__version_info__'
]

__title__ = 'google-music-utils'
__summary__ = 'A set of utility functionality for google-music and related projects.'
__url__ = 'https://github.com/thebigmunch/google-music-utils'

__version__ = '1.0.0'
__version_info__ = tuple(int(i) for i in __version__.split('.') if i.isdigit())

__author__ = 'thebigmunch'
__author_email__ = 'mail@thebigmunch.me'

__license__ = 'MIT'
__copyright__ = f'2018 {__author__} <{__author_email__}>'



__all__ = [
	'gm_timestamp', 'from_gm_timestamp', 'to_gm_timestamp',
	'is_album_id', 'is_artist_id', 'is_podcast_episode_id',
	'is_podcast_series_id', 'is_store_song_id', 'is_uuid'
]

import re
import time

_uuid_re = re.compile(r'^[0-9a-fA-F]{8}-([0-9a-fA-F]{4}-){3}[0-9a-fA-F]{12}$')


def gm_timestamp():
	"""Generate a timestamp in microseconds."""

	return int(time.time() * 1000000)


def from_gm_timestamp(timestamp):
	"""Convert timestamp in microseconds to timestamp in seconds."""

	return int(timestamp / 1000000)


def to_gm_timestamp(timestamp):
	"""Convert timestamp in seconds to timestamp in microseconds."""

	return int(timestamp * 1000000)


def is_album_id(item_id):
	"""Validate if ID is in the format of a Google Music album ID."""

	return len(item_id) == 27 and item_id.startswith('B')


def is_artist_id(item_id):
	"""Validate if ID is in the format of a Google Music artist ID."""

	return len(item_id) == 27 and item_id.startswith('A')


def is_podcast_episode_id(item_id):
	"""Validate if ID is in the format of a Google Music podcast episode ID."""

	return len(item_id) == 27 and item_id.startswith('D')


def is_podcast_series_id(item_id):
	"""Validate if ID is in the format of a Google Music series ID."""

	return len(item_id) == 27 and item_id.startswith('I')


def is_store_song_id(item_id):
	"""Validate if ID is in the format of a Google Music store song ID."""

	return len(item_id) == 27 and item_id.startswith('T')


def is_uuid(item_id):
	"""Validate if ID is in the format of a UUID (used for library song IDs)."""

	return bool(_uuid_re.match(item_id))



__all__ = ['list_to_single_value', 'normalize_value']

import re


def list_to_single_value(value):
	if not isinstance(value, (list, tuple)):
		return value

	return value[0]


def normalize_value(value):
	"""Normalize metadata value to improve match accuracy."""

	value = str(value)
	value = value.lower()

	value = re.sub(r'\/\s*\d+', '', value)  # Remove "/<totaltracks>" from track number.
	value = re.sub(r'^0+([0-9]+)', r'\1', value)  # Remove leading zero(s) from track number.
	value = re.sub(r'^(\d+)\.+', r'\1', value)  # Remove dots from track number.
	value = re.sub(r'[^\w\s]', '', value)  # Remove leading non-word characters.
	value = re.sub(r'^the\s+', '', value)  # Remove leading "the".
	value = re.sub(r'^\s+', '', value)  # Remove leading spaces.
	value = re.sub(r'\s+$', '', value)  # Remove trailing spaces.
	value = re.sub(r'\s+', ' ', value)  # Reduce multiple spaces to a single space.

	return value



__all__ = ['CHARACTER_REPLACEMENTS', 'FIELD_MAP', 'TEMPLATE_PATTERNS']

from multidict import MultiDict

CHARACTER_REPLACEMENTS = {
	'\\': '-', '/': ',', ':': '-', '*': 'x', '<': '[',
	'>': ']', '|': '!', '?': '', '"': "''"
}
"""dict: Mapping of invalid filepath characters with appropriate replacements."""

_FIELD_MAP_GROUPS = [
	('albumartist', 'album_artist', 'albumArtist'), ('bpm', 'beatsPerMinute'), ('date', 'year'),
	('discnumber', 'disc_number', 'discNumber'), ('disctotal', 'total_disc_count', 'totalDiscCount'),
	('tracknumber', 'track_number', 'trackNumber'), ('tracktotal', 'total_track_count', 'totalTrackCount')
]
_FIELD_MAP = [(field, alias) for group in _FIELD_MAP_GROUPS for field in group for alias in group if field != alias]

FIELD_MAP = MultiDict(_FIELD_MAP)

# TODO: Support other/more metadata field names.
# TODO: Support multiple field names per template pattern.
TEMPLATE_PATTERNS = {
	'%artist%': 'artist', '%title%': 'title', '%tracknumber%': 'tracknumber',
	'%album%': 'album', '%date%': 'date', '%genre%': 'genre',
	'%albumartist%': 'albumartist', '%discnumber%': 'discnumber'
}
"""dict: Mapping of template patterns to their audio-metadata field name."""



__all__ = ['exclude_items', 'include_items']

import functools
import re
from itertools import filterfalse

import audio_metadata

from .utils import normalize_value


def _match_field(field_value, pattern, ignore_case=False, normalize_values=False):
	"""Match an item metadata field value by pattern.

	Note:
		Metadata values are lowercased when ``normalized_values`` is ``True``,
		so ``ignore_case`` is automatically set to ``True``.

	Parameters:
		field_value (list or str): A metadata field value to check.
		pattern (str): A regex pattern to check the field value(s) against.
		ignore_case (bool): Perform case-insensitive matching.
			Default: ``False``
		normalize_values (bool): Normalize metadata values to remove common differences between sources.
			Default: ``False``

	Returns:
		bool: True if matched, False if not.
	"""

	if normalize_values:  # pragma: no cover
		ignore_case = True

	normalize = normalize_value if normalize_values else lambda x: str(x)
	search = functools.partial(re.search, flags=re.I) if ignore_case else re.search

	# audio_metadata fields contain a list of values.
	if isinstance(field_value, list):
		return any(search(pattern, normalize(value)) for value in field_value)
	else:
		return search(pattern, normalize(field_value))


def _match_item(item, any_all=any, ignore_case=False, normalize_values=False, **kwargs):
	"""Match items by metadata.

	Note:
		Metadata values are lowercased when ``normalized_values`` is ``True``,
		so ``ignore_case`` is automatically set to ``True``.

	Parameters:
		item (dict): Item dict or filepath.
		any_all (callable): A callable to determine if any or all filters must match to match item.
			Expected values :obj:`any` (default) or :obj:`all`.
		ignore_case (bool): Perform case-insensitive matching.
			Default: ``False``
		normalize_values (bool): Normalize metadata values to remove common differences between sources.
			Default: ``False``
		kwargs (list): Lists of values to match the given metadata field.

	Returns:
		bool: True if matched, False if not.
	"""

	if not isinstance(item, (audio_metadata.Format, dict)):
		it = audio_metadata.load(item).tags  # pragma: no cover
	else:
		it = item

	return any_all(
		_match_field(
			it.get(field, ''), pattern, ignore_case=ignore_case, normalize_values=normalize_values
		) for field, patterns in kwargs.items() for pattern in patterns
	)


def exclude_items(items, any_all=any, ignore_case=False, normalize_values=False, **kwargs):
	"""Exclude items by matching metadata.

	Note:
		Metadata values are lowercased when ``normalized_values`` is ``True``,
		so ``ignore_case`` is automatically set to ``True``.

	Parameters:
		items (list): A sequence of item dicts or filepaths.
		any_all (callable): A callable to determine if any or all filters must match to exclude items.
			Expected values :obj:`any` (default) or :obj:`all`.
		ignore_case (bool): Perform case-insensitive matching.
			Default: ``False``
		normalize_values (bool): Normalize metadata values to remove common differences between sources.
			Default: ``False``
		kwargs (list): Sequences of values to match the given metadata field.

	Yields:
		dict: The next item to be included.

	Example:
		>>> from google_music_utils import exclude_items
		>>> list(exclude_items(song_list, any_all=all, ignore_case=True, normalize_values=True, artist=['Beck'], album=['Golden Feelings']))
	"""

	if kwargs:
		match = functools.partial(
			_match_item, any_all=any_all, ignore_case=ignore_case, normalize_values=normalize_values, **kwargs
		)

		return filterfalse(match, items)
	else:
		return iter(items)


def include_items(items, any_all=any, ignore_case=False, normalize_values=False, **kwargs):
	"""Include items by matching metadata.

	Note:
		Metadata values are lowercased when ``normalized_values`` is ``True``,
		so ``ignore_case`` is automatically set to ``True``.

	Parameters:
		items (list): A sequence of item dicts or filepaths.
		any_all (callable): A callable to determine if any or all filters must match to include items.
			Expected values :obj:`any` (default) or :obj:`all`.
		ignore_case (bool): Perform case-insensitive matching.
			Default: ``False``
		normalize_values (bool): Normalize metadata values to remove common differences between sources.
			Default: ``False``
		kwargs (list): Sequences of values to match the given metadata field.

	Yields:
		dict: The next item to be included.

	Example:
		>>> from google_music_utils import exclude_items
		>>> list(include_items(song_list, any_all=all, ignore_case=True, normalize_values=True, artist=['Beck'], album=['Odelay']))
	"""

	if kwargs:
		match = functools.partial(
			_match_item, any_all=any_all, ignore_case=ignore_case, normalize_values=normalize_values, **kwargs
		)

		return filter(match, items)
	else:
		return iter(items)



# -*- coding: utf-8 -*-
from distutils.core import setup

package_dir = \
{'': 'src'}

packages = \
['google_music_utils']

package_data = \
{'': ['*']}

install_requires = \
['audio-metadata>=0.1,<0.2', 'multidict>=4.0,<5.0', 'wrapt>=1.10,<2.0']

extras_require = \
{'dev': ['coverage>=4.5,<5.0',
         'flake8>=3.5,<4.0',
         'flake8-builtins>=1.0,<2.0',
         'flake8-import-order>=0.18,<0.19',
         'flake8-import-order-tbm>=1.0.0,<2.0.0',
         'pytest>=3.6,<4.0',
         'sphinx>=1.7,<2.0',
         'tox>=2.9,<3.0'],
 'doc': ['sphinx>=1.7,<2.0'],
 'lint': ['flake8>=3.5,<4.0',
          'flake8-builtins>=1.0,<2.0',
          'flake8-import-order>=0.18,<0.19',
          'flake8-import-order-tbm>=1.0.0,<2.0.0'],
 'test': ['coverage>=4.5,<5.0', 'pytest>=3.6,<4.0', 'tox>=2.9,<3.0']}

setup_kwargs = {
    'name': 'google-music-utils',
    'version': '1.0.0',
    'description': 'A set of utility functionality for google-music and related projects.',
    'long_description': '# google-music-utils\n\n[google-music-utils](https://github.com/thebigmunch/google-music-utils) is a\nset of utility functionality for google-music and related projects.\n\n\n## Installation\n\n``pip install google-music-utils``\n\n\n## Usage\n\nSee the [docs](https://google-music-utils.readthedocs.io).\n\n\n## Appreciation\n\nShowing appreciation is always welcome.\n\n#### Thank\n\n[![Say Thanks](https://img.shields.io/badge/thank-thebigmunch-blue.svg?style=flat-square)](https://saythanks.io/to/thebigmunch)\n\nGet your own thanks inbox at [SayThanks.io](https://saythanks.io/).\n\n#### Contribute\n\n[Contribute](https://github.com/thebigmunch/google-music-utils/blob/master/CONTRIBUTING.md) by submitting bug reports, feature requests, or code.\n\n#### Help Others/Stay Informed\n\n[Discourse forum](https://forum.thebigmunch.me/)\n\n#### Referrals/Donations\n\n[![Coinbase](https://img.shields.io/badge/Coinbase-referral-orange.svg?style=flat-square)](https://www.coinbase.com/join/52502f01e0fdd4d3ef000253) [![Digital Ocean](https://img.shields.io/badge/Digital_Ocean-referral-orange.svg?style=flat-square)](https://m.do.co/c/3823208a0597) [![Namecheap](https://img.shields.io/badge/Namecheap-referral-orange.svg?style=flat-square)](https://www.namecheap.com/?aff=67208) [![PayPal](https://img.shields.io/badge/PayPal-donate-brightgreen.svg?style=flat-square)](https://www.paypal.com/cgi-bin/webscr?cmd=_donations&business=DHDVLSYW8V8N4&lc=US&item_name=thebigmunch&currency_code=USD)\n\n**BTC:** ``1BMLCFPcX8YHE1He2t3aBrsNDGr1pKhfFa``  \n**ETH:** ``0x8E3f8d8eAedeA61Bf34A998A2104954FE508D5d0``  \n**LTC:** ``LgsQU1YaY4a4s7m9efjn6m35XhVEpW1xoP``\n',
    'author': 'thebigmunch',
    'author_email': 'mail@thebigmunch.me',
    'url': 'https://github.com/thebigmunch/google-music-utils',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
